@@ -9,6 +9,12 @@ import shutil
 from pathlib import Path
 from typing import Optional
 
+class DockerException(Exception):
+    def __init__(self, message, stdout=None, stderr=None):
+        super().__init__(message)
+        self.stdout = stdout
+        self.stderr = stderr
+
 class DockerComposeClient:
     def __init__(self, project=None, services=[], compose_file='docker-compose.yml'):
         self._services = services
@@ -26,6 +32,23 @@ class DockerComposeClient:
         else:
             return run_compose_cmd(*self._pre, cmd, *args, *self._services, *self._post, **kwargs)
 
+def get_docker_runtime(desired: str) -> str:
+    """Checks dockers available runtimes and will return desired if it's available
+
+    Args:
+        desired (str): The desired runtime, will return this if it's available
+
+    Returns:
+        str: Either the desired or the default docker runtime if desired is not an actual runtime
+    """
+    import re
+
+    stdout, _ = run_docker_cmd("--debug", "info", stdout=-1)
+    if re.search(f"Runtimes:.*{desired}", stdout) is not None:
+        return desired
+    else:
+        return re.search("Default Runtime: (.*)", stdout).group(1)
+
 def get_docker_client_binary_path() -> Optional[Path]:
     """Return the path of the docker client binary file.
 
@@ -47,7 +70,7 @@ def compose_is_installed() -> bool:
     Returns:
         bool: whether docker compose (v2) is installed.
     """
-    help_output = run_docker_cmd("compose", "--help", stdout=subprocess.PIPE)
+    help_output, _ = run_docker_cmd("compose", "--help", stdout=subprocess.PIPE)
     return "compose" in help_output
 
 def run_compose_cmd(*args, **kwargs):
@@ -58,10 +81,11 @@ def run_docker_cmd(*args, **kwargs):
     """
 
     docker_binary = get_docker_client_binary_path()
-    return _run([docker_binary, *args], **kwargs)
+    return _run(*[docker_binary, *args], **kwargs)
 
 def _run(*args, **kwargs):
-    LOGGER.info(f"{' '.join([str(arg) for arg in args[0]])}")
+    cmd = ' '.join([str(arg) for arg in args])
+    LOGGER.info(f"{cmd}")
 
     def post_process_stream(stream: Optional[bytes]):
         if stream is None:
@@ -71,8 +95,16 @@ def _run(*args, **kwargs):
             stream = stream[:-1]
         return stream
 
-    completed_process = subprocess.run(*args, **kwargs)
-    return post_process_stream(completed_process.stdout)
+    args = [arg for arg in args if arg]
+    completed_process = subprocess.run(args, **kwargs)
+
+    stdout = post_process_stream(completed_process.stdout)
+    stderr = post_process_stream(completed_process.stderr)
+    
+    if completed_process.returncode:
+        raise DockerException(f"Got an error code of '{completed_process.returncode}': {cmd}", stdout, stderr)
+
+    return stdout, stderr
 
 # Ports
 # From https://github.com/containers/podman-compose/blob/devel/podman_compose.py
@@ -121,3 +153,33 @@ def find_available_port(port, trys=5):
         else:
             break
     return port if not in_use else None
+
+# Devices
+def parse_devices(devices):
+    device_list = []
+    for device in devices:
+        if isinstance(device, dict):
+            device_list.append(device)
+            continue
+        if not isinstance(device, str):
+            raise errors.DockerException(
+                f'Invalid device type {type(device)}'
+            )
+        device_mapping = device.split(':')
+        if device_mapping:
+            path_on_host = device_mapping[0]
+            if len(device_mapping) > 1:
+                path_in_container = device_mapping[1]
+            else:
+                path_in_container = path_on_host
+            if len(device_mapping) > 2:
+                permissions = device_mapping[2]
+            else:
+                permissions = 'rwm'
+            device_list.append({
+                'PathOnHost': path_on_host,
+                'PathInContainer': path_in_container,
+                'CgroupPermissions': permissions,
+                'Original': device
+            })
+    return device_list
